import os
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence

from jina.logging.logger import JinaLogger
from jina.serve.networking.instrumentation import (
    _NetworkingHistograms,
    _NetworkingMetrics,
)
from jina.serve.networking.replica_list import _ReplicaList

if TYPE_CHECKING:  # pragma: no cover

    from grpc.aio._interceptor import ClientInterceptor
    from opentelemetry.instrumentation.grpc._client import (
        OpenTelemetryClientInterceptor,
    )


class _ConnectionPoolMap:
    def __init__(
        self,
        runtime_name: str,
        logger: Optional[JinaLogger],
        metrics: _NetworkingMetrics,
        histograms: _NetworkingHistograms,
        aio_tracing_client_interceptors: Optional[Sequence['ClientInterceptor']] = None,
        tracing_client_interceptor: Optional['OpenTelemetryClientInterceptor'] = None,
        channel_options: Optional[list] = None,
    ):
        self._logger = logger
        # this maps deployments to shards or heads
        self._deployments: Dict[str, Dict[str, Dict[int, _ReplicaList]]] = {}
        # dict stores last entity id used for a particular deployment, used for round robin
        self._access_count: Dict[str, int] = {}
        self._metrics = metrics
        self._histograms = histograms
        self.runtime_name = runtime_name
        if os.name != 'nt':
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')
        self.aio_tracing_client_interceptors = aio_tracing_client_interceptors
        self.tracing_client_interceptor = tracing_client_interceptor
        self.channel_options = channel_options

    def add_replica(self, deployment: str, shard_id: int, address: str):
        self._add_connection(deployment, shard_id, address, 'shards')

    def add_head(
        self, deployment: str, address: str, head_id: Optional[int] = 0
    ):  # the head_id is always 0 for now, this will change when scaling the head
        self._add_connection(deployment, head_id, address, 'heads')

    def get_replicas(
        self,
        deployment: str,
        head: bool,
        entity_id: Optional[int] = None,
        increase_access_count: bool = True,
    ) -> Optional[_ReplicaList]:
        # returns all replicas of a given deployment, using a given shard
        if deployment in self._deployments:
            type_ = 'heads' if head else 'shards'
            if entity_id is None and head:
                entity_id = 0
            return self._get_connection_list(
                deployment, type_, entity_id, increase_access_count
            )
        else:
            self._logger.debug(
                f'Unknown deployment {deployment}, no replicas available'
            )
            return None

    def get_replicas_all_shards(self, deployment: str) -> List[_ReplicaList]:
        # returns all replicas of a given deployment, for all available shards
        # result is a list of 'shape' (num_shards, num_replicas), containing all replicas for all shards
        replicas = []
        if deployment in self._deployments:
            for shard_id in self._deployments[deployment]['shards']:
                replicas.append(
                    self._get_connection_list(deployment, 'shards', shard_id)
                )
        return replicas

    async def close(self):
        # Close all connections to all replicas
        for deployment in self._deployments:
            for entity_type in self._deployments[deployment]:
                for shard_in in self._deployments[deployment][entity_type]:
                    await self._deployments[deployment][entity_type][shard_in].close()
        self._deployments.clear()

    def _get_connection_list(
        self,
        deployment: str,
        type_: str,
        entity_id: Optional[int] = None,
        increase_access_count: bool = True,
    ) -> Optional[_ReplicaList]:
        try:
            if entity_id is None and len(self._deployments[deployment][type_]) > 0:
                # select a random entity
                if increase_access_count:
                    self._access_count[deployment] += 1
                return self._deployments[deployment][type_][
                    self._access_count[deployment]
                    % len(self._deployments[deployment][type_])
                ]
            else:
                return self._deployments[deployment][type_][entity_id]
        except KeyError:
            if (
                entity_id is None
                and deployment in self._deployments
                and len(self._deployments[deployment][type_])
            ):
                # This can happen as a race condition when removing connections while accessing it
                # In this case we don't care for the concrete entity, so retry with the first one
                return self._get_connection_list(
                    deployment, type_, 0, increase_access_count
                )
            return None

    def _add_deployment(self, deployment: str):
        if deployment not in self._deployments:
            self._deployments[deployment] = {'shards': {}, 'heads': {}}
            self._access_count[deployment] = 0

    def _add_connection(
        self,
        deployment: str,
        entity_id: int,
        address: str,
        type: str,
    ):
        self._add_deployment(deployment)
        if entity_id not in self._deployments[deployment][type]:
            connection_list = _ReplicaList(
                metrics=self._metrics,
                histograms=self._histograms,
                logger=self._logger,
                runtime_name=self.runtime_name,
                aio_tracing_client_interceptors=self.aio_tracing_client_interceptors,
                tracing_client_interceptor=self.tracing_client_interceptor,
                deployment_name=deployment,
                channel_options=self.channel_options,
            )
            self._deployments[deployment][type][entity_id] = connection_list

        if not self._deployments[deployment][type][entity_id].has_connection(address):
            self._logger.debug(
                f'adding connection for deployment {deployment}/{type}/{entity_id} to {address}'
            )
            self._deployments[deployment][type][entity_id].add_connection(
                address, deployment_name=deployment
            )
            self._logger.debug(
                f'connection for deployment {deployment}/{type}/{entity_id} to {address} added'
            )
        else:
            self._logger.debug(
                f'ignoring activation of pod for deployment {deployment}, {address} already known'
            )

    async def remove_head(self, deployment, address, head_id: Optional[int] = 0):
        return await self._remove_connection(deployment, head_id, address, 'heads')

    async def remove_replica(self, deployment, address, shard_id: Optional[int] = 0):
        return await self._remove_connection(deployment, shard_id, address, 'shards')

    async def _remove_connection(self, deployment, entity_id, address, type):
        if (
            deployment in self._deployments
            and entity_id in self._deployments[deployment][type]
        ):
            self._logger.debug(
                f'removing connection for deployment {deployment}/{type}/{entity_id} to {address}'
            )
            await self._deployments[deployment][type][entity_id].remove_connection(
                address
            )
            if not self._deployments[deployment][type][entity_id].has_connections():
                self._deployments[deployment][type].pop(entity_id)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import add_voter_pb2 as add__voter__pb2


class RaftAdminStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddVoter = channel.unary_unary(
                '/RaftAdmin/AddVoter',
                request_serializer=add__voter__pb2.AddVoterRequest.SerializeToString,
                response_deserializer=add__voter__pb2.Future.FromString,
                )
        self.Await = channel.unary_unary(
                '/RaftAdmin/Await',
                request_serializer=add__voter__pb2.Future.SerializeToString,
                response_deserializer=add__voter__pb2.AwaitResponse.FromString,
                )
        self.Forget = channel.unary_unary(
                '/RaftAdmin/Forget',
                request_serializer=add__voter__pb2.Future.SerializeToString,
                response_deserializer=add__voter__pb2.ForgetResponse.FromString,
                )


class RaftAdminServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AddVoter(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Await(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Forget(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RaftAdminServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddVoter': grpc.unary_unary_rpc_method_handler(
                    servicer.AddVoter,
                    request_deserializer=add__voter__pb2.AddVoterRequest.FromString,
                    response_serializer=add__voter__pb2.Future.SerializeToString,
            ),
            'Await': grpc.unary_unary_rpc_method_handler(
                    servicer.Await,
                    request_deserializer=add__voter__pb2.Future.FromString,
                    response_serializer=add__voter__pb2.AwaitResponse.SerializeToString,
            ),
            'Forget': grpc.unary_unary_rpc_method_handler(
                    servicer.Forget,
                    request_deserializer=add__voter__pb2.Future.FromString,
                    response_serializer=add__voter__pb2.ForgetResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'RaftAdmin', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RaftAdmin(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AddVoter(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RaftAdmin/AddVoter',
            add__voter__pb2.AddVoterRequest.SerializeToString,
            add__voter__pb2.Future.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Await(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RaftAdmin/Await',
            add__voter__pb2.Future.SerializeToString,
            add__voter__pb2.AwaitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Forget(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RaftAdmin/Forget',
            add__voter__pb2.Future.SerializeToString,
            add__voter__pb2.ForgetResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import asyncio
import json
from contextlib import AsyncExitStack
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Type

from jina._docarray import Document, DocumentArray, docarray_v2
from jina.clients.base import BaseClient
from jina.clients.base.helper import HTTPClientlet, handle_response_status
from jina.clients.helper import callback_exec
from jina.importer import ImportExtensions
from jina.logging.profile import ProgressBar
from jina.serve.stream import RequestStreamer
from jina.types.request import Request
from jina.types.request.data import DataRequest

if TYPE_CHECKING:  # pragma: no cover
    from jina.clients.base import CallbackFnType, InputType


class HTTPBaseClient(BaseClient):
    """A MixIn for HTTP Client."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._endpoints = []

    async def _get_endpoints_from_openapi(self):
        def extract_paths_by_method(spec):
            paths_by_method = {}
            for path, methods in spec['paths'].items():
                for method, details in methods.items():
                    if method not in paths_by_method:
                        paths_by_method[method] = []
                    paths_by_method[method].append(path.strip('/'))

            return paths_by_method

        import json

        import aiohttp

        proto = 'https' if self.args.tls else 'http'
        target_url = f'{proto}://{self.args.host}:{self.args.port}/openapi.json'
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(target_url) as response:
                    content = await response.read()
                    openapi_response = json.loads(content.decode())
                    self._endpoints = extract_paths_by_method(openapi_response).get(
                        'post', []
                    )
        except:
            pass

    async def _is_flow_ready(self, **kwargs) -> bool:
        """Sends a dry run to the Flow to validate if the Flow is ready to receive requests

        :param kwargs: kwargs coming from the public interface. Includes arguments to be passed to the `HTTPClientlet`
        :return: boolean indicating the health/readiness of the Flow
        """
        from jina.proto import jina_pb2

        async with AsyncExitStack() as stack:
            try:
                proto = 'https' if self.args.tls else 'http'
                url = f'{proto}://{self.args.host}:{self.args.port}/dry_run'
                iolet = await stack.enter_async_context(
                    HTTPClientlet(
                        url=url,
                        logger=self.logger,
                        tracer_provider=self.tracer_provider,
                        **kwargs,
                    )
                )

                response = await iolet.send_dry_run(**kwargs)
                r_status = response.status

                r_str = await response.json()
                handle_response_status(r_status, r_str, url)
                if r_str['code'] == jina_pb2.StatusProto.SUCCESS:
                    return True
                else:
                    self.logger.error(
                        f'Returned code is not expected! Description: {r_str["description"]}'
                    )
            except Exception as e:
                self.logger.error(
                    f'Error while fetching response from HTTP server {e!r}'
                )
        return False

    async def _get_results(
        self,
        inputs: 'InputType',
        on_done: 'CallbackFnType',
        on_error: Optional['CallbackFnType'] = None,
        on_always: Optional['CallbackFnType'] = None,
        max_attempts: int = 1,
        initial_backoff: float = 0.5,
        max_backoff: float = 0.1,
        backoff_multiplier: float = 1.5,
        results_in_order: bool = False,
        prefetch: Optional[int] = None,
        timeout: Optional[int] = None,
        return_type: Type[DocumentArray] = DocumentArray,
        **kwargs,
    ):
        """
        :param inputs: the callable
        :param on_done: the callback for on_done
        :param on_error: the callback for on_error
        :param on_always: the callback for on_always
        :param max_attempts: Number of sending attempts, including the original request.
        :param initial_backoff: The first retry will happen with a delay of random(0, initial_backoff)
        :param max_backoff: The maximum accepted backoff after the exponential incremental delay
        :param backoff_multiplier: The n-th attempt will occur at random(0, min(initialBackoff*backoffMultiplier**(n-1), maxBackoff))
        :param results_in_order: return the results in the same order as the inputs
        :param prefetch: How many Requests are processed from the Client at the same time.
        :param timeout: Timeout for the client to remain connected to the server.
        :param return_type: the DocumentArray type to be returned. By default, it is `DocumentArray`.
        :param kwargs: kwargs coming from the public interface. Includes arguments to be passed to the `HTTPClientlet`
        :yields: generator over results
        """
        with ImportExtensions(required=True):
            pass

        self.inputs = inputs
        request_iterator = self._get_requests(**kwargs)
        on = kwargs.get('on', '/post')
        if len(self._endpoints) == 0:
            await self._get_endpoints_from_openapi()

        async with AsyncExitStack() as stack:
            cm1 = ProgressBar(
                total_length=self._inputs_length, disable=not self.show_progress
            )
            p_bar = stack.enter_context(cm1)
            proto = 'https' if self.args.tls else 'http'
            endpoint = on.strip('/')
            has_default_endpoint = 'default' in self._endpoints

            if endpoint != '' and endpoint in self._endpoints:
                url = f'{proto}://{self.args.host}:{self.args.port}/{on.strip("/")}'
            elif has_default_endpoint:
                url = f'{proto}://{self.args.host}:{self.args.port}/default'
            else:
                url = f'{proto}://{self.args.host}:{self.args.port}/post'

            iolet = await stack.enter_async_context(
                HTTPClientlet(
                    url=url,
                    logger=self.logger,
                    tracer_provider=self.tracer_provider,
                    max_attempts=max_attempts,
                    initial_backoff=initial_backoff,
                    max_backoff=max_backoff,
                    backoff_multiplier=backoff_multiplier,
                    timeout=timeout,
                    **kwargs,
                )
            )

            def _request_handler(
                request: 'Request', **kwargs
            ) -> 'Tuple[asyncio.Future, Optional[asyncio.Future]]':
                """
                For HTTP Client, for each request in the iterator, we `send_message` using
                http POST request and add it to the list of tasks which is awaited and yielded.
                :param request: current request in the iterator
                :param kwargs: kwargs
                :return: asyncio Task for sending message
                """
                return asyncio.ensure_future(iolet.send_message(request=request)), None

            def _result_handler(result):
                return result

            streamer_args = vars(self.args)
            if prefetch:
                streamer_args['prefetch'] = prefetch
            streamer = RequestStreamer(
                request_handler=_request_handler,
                result_handler=_result_handler,
                logger=self.logger,
                **streamer_args,
            )
            async for response in streamer.stream(
                request_iterator=request_iterator, results_in_order=results_in_order
            ):
                r_status = response.status

                r_str = await response.json()
                handle_response_status(r_status, r_str, url)

                da = None
                if 'data' in r_str and r_str['data'] is not None:
                    from jina._docarray import DocumentArray, docarray_v2

                    if not docarray_v2:
                        da = DocumentArray.from_dict(r_str['data'])
                    else:
                        from docarray import DocList

                        if issubclass(return_type, DocList):
                            da = return_type(
                                [return_type.doc_type(**v) for v in r_str['data']]
                            )
                        else:
                            da = DocList[return_type](
                                [return_type(**v) for v in r_str['data']]
                            )
                    del r_str['data']

                resp = DataRequest(r_str)
                if da is not None:
                    resp.data.docs = da

                callback_exec(
                    response=resp,
                    logger=self.logger,
                    on_error=on_error,
                    on_done=on_done,
                    on_always=on_always,
                    continue_on_error=self.continue_on_error,
                )
                if self.show_progress:
                    p_bar.update()
                yield resp

    async def _get_streaming_results(
        self,
        on: str,
        inputs: 'Document',
        parameters: Optional[Dict] = None,
        return_type: Type[Document] = Document,
        timeout: Optional[int] = None,
        **kwargs,
    ):
        proto = 'https' if self.args.tls else 'http'
        endpoint = on.strip('/')
        has_default_endpoint = 'default' in self._endpoints

        if (endpoint != '' and endpoint in self._endpoints) or not has_default_endpoint:
            url = f'{proto}://{self.args.host}:{self.args.port}/{endpoint}'
        else:
            url = f'{proto}://{self.args.host}:{self.args.port}/default'

        iolet = HTTPClientlet(
            url=url,
            logger=self.logger,
            tracer_provider=self.tracer_provider,
            timeout=timeout,
            **kwargs,
        )

        async with iolet:
            async for doc in iolet.send_streaming_message(doc=inputs, on=on):
                if not docarray_v2:
                    yield Document.from_dict(json.loads(doc))
                else:
                    yield return_type(**json.loads(doc))
